/*
    Copyright 2005-2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "namedparametertreeitem.h"
#include "namedparametertreeitem.moc"

#include <QtCore/QDebug>

namespace NSCore {

	NamedParameterTreeItem::NamedParameterTreeItem( QString n, QVariant v, QTreeItem* p )
		: QTreeItem( n, p )
		, _value( v )
		, _min( QVariant::Invalid )
		, _max( QVariant::Invalid )
		, _inupdate( false )
		, _guihint( 3 )
	{ }
	NamedParameterTreeItem::NamedParameterTreeItem( QString n, QVariant::Type t, QTreeItem* p )
		: QTreeItem( n, p )
		, _value( QVariant( t ) )
		, _min( QVariant::Invalid )
		, _max( QVariant::Invalid )
		, _inupdate( false )
		, _guihint( 3 )
	{ }

	NamedParameterTreeItem::NamedParameterTreeItem( NamedParameterTreeItem& item, QTreeItem* p )
		: QTreeItem( item.objectName(), p )
	{
		foreach( QString s, item.childlist() ) {
			NamedParameterTreeItem* tmp = dynamic_cast<NamedParameterTreeItem*>( item.child( s ) );
			if ( tmp != 0 )
				new NamedParameterTreeItem( *tmp, this );
		}
	}
	NamedParameterTreeItem::~NamedParameterTreeItem() {
		//qDebug() << "NamedParameterTreeItem::~NamedParameterTreeItem()" << path() << " " << objectName();
	}

	void NamedParameterTreeItem::setValue( QVariant v ) {
		//qDebug() << "NamedParameterTreeItem::setValue(" << v << ")";
		//qDebug() << " _inupdate=" << _inupdate << " canConvert()" << v.canConvert( _value.type() );

		if ( !_inupdate && v.canConvert( _value.type() ) )
		{
			if ( _min.type() != QVariant::Invalid && _max.type() != QVariant::Invalid ) {
				if (
					( _value.type()==QVariant::Double && v.toDouble() >= _min.toDouble() && v.toDouble() <= _max.toDouble() )
					||
					( _value.type()==QVariant::Int && v.toInt() >= _min.toInt() && v.toInt() <= _max.toInt() )
					||
					( _value.type() != QVariant::Double && _value.type() != QVariant::Int )
				)
				{
					//qDebug() << " Boundaries are checked and okay";
				} else {
					// Missed boundaries
					qWarning() << "NamedParameterTreeItem::setValue(" << v << ")";
					qWarning() << QString( " Boundaries missed! [%1,%2]" ).arg( _min.toString() ).arg( _max.toString() );
					return;
				}
			}// else
			//	qDebug() << " No boundaries set";

			_inupdate = true; // Begin with blocking the update for other calls

			v.convert( _value.type() );

			_value = v;

			emit valueChanged();

			QObject* _sender = sender();
			//qDebug() << "Parameter::setValue(" << v << ") sender =" << _sender;

			if ( receivers( SIGNAL( valueChanged( QVariant ) ) ) == 1 && _sender != 0 ) {
				emit valueChanged( _value );
			}
	
			if ( _sender == 0 || receivers( SIGNAL( valueChanged( int ) ) ) > 1 )
				if ( _value.canConvert( QVariant::Int ) ) emit valueChanged( _value.toInt() );
			if ( _sender == 0 || receivers( SIGNAL( valueChanged( bool ) ) ) > 1 )
				if ( _value.canConvert( QVariant::Bool ) ) emit valueChanged( _value.toBool() );
			if ( _sender == 0 || receivers( SIGNAL( valueChanged( double ) ) ) > 1 )
				if ( _value.canConvert( QVariant::Double ) ) emit valueChanged( _value.toDouble() );
			if ( _sender == 0 || receivers( SIGNAL( valueChanged( QString ) ) ) > 1 )
				if ( _value.canConvert( QVariant::String ) ) emit valueChanged( _value.toString() );

			_inupdate = false; // Release the lock
			//QTimer::singleShot( 20, this, SLOT( unlockUpdate() ) );
		}
		//else
		//	qWarning() << "Parameter::setValue( QVariant ) Either in update or wrong type (" << _value.type() << "|" << v.type() << "), _inupdate =" << _inupdate;
	}

	void NamedParameterTreeItem::setValue( int n ) { setValue( QVariant( n ) ); }
	void NamedParameterTreeItem::setValue( bool n ) { setValue( QVariant( n ) ); }
	void NamedParameterTreeItem::setValue( double n ) { setValue( QVariant( n ) ); }
	void NamedParameterTreeItem::setValue( QString n ) { setValue( QVariant( n ) ); }

	void NamedParameterTreeItem::setMin( QVariant v ) { _min = v; }
	void NamedParameterTreeItem::setMax( QVariant v ) { _max = v; }

	void NamedParameterTreeItem::reset() {
		setValue( QVariant( _value.type() ) );
	}

	void NamedParameterTreeItem::unlockUpdate() {
		_inupdate = false;
	}

	NamedParameterTreeItem* NamedParameterTreeItem::getParameter( QString path ) {
		//qDebug() << "NamedParameterTreeItem::getParameter(" << path << ")";
		QTreeItem* item = getNamedChild( path );
		NamedParameterTreeItem* parameter = dynamic_cast<NamedParameterTreeItem*>( item );
		if ( parameter ) {
			//qDebug() << " found a valid Parameter" << parameter;
			return parameter;
		}
		qWarning() << "NamedParameterTreeItem::getParameter(" << path << ")\n  Will return 0!";
		return 0;
	}

	void NamedParameterTreeItem::addParameter( QString path, const NamedParameterTreeItem* parameter, QString tooltip ) {
		qWarning() << "NamedParameterTreeItem::addParameter(" << path << "," << parameter << "," << tooltip << ") Is deprecated and should not be called anymore!";
	}
	void NamedParameterTreeItem::addParameter( QString path, QVariant value, QString tooltip ) {
		//qDebug() << "NamedParameterTreeItem::addParameter(" << path << "," << value << ")";
		QString parentpath = path.section( "/", 0, -2 );
		QTreeItem* p = getNamedChild( parentpath );
		if ( parentpath != "" && p == 0 ) {
			QString firstpath = parentpath.section( "/", 0, 0 );
			NamedParameterTreeItem* item = dynamic_cast<NamedParameterTreeItem*>( getParameter( firstpath ) );
			if ( !item )
				addParameter( firstpath, QVariant::Invalid );
			item = dynamic_cast<NamedParameterTreeItem*>( getParameter( firstpath ) );
			if ( item )
				item->addParameter( path.section( "/", 1, -1 ), value );
			else
				qWarning() << "! Warning: couldn't create child!";
		} else {
			if ( parentpath == "" )
				p = this;
			NamedParameterTreeItem* tmp = new NamedParameterTreeItem( path.section( "/", -1, -1 ), value, p );
			connect( tmp, SIGNAL( valueChanged() ), dynamic_cast<NamedParameterTreeItem*>( p ), SIGNAL( valueChanged() ) );
			tmp->setToolTip( tooltip );
		}
	}
	void NamedParameterTreeItem::addParameter( QString path, QVariant::Type type, QString tooltip ) {
		//qDebug() << "NamedParameterTreeItem::addParameter(" << path << "," << type << ")";
		addParameter( path, QVariant( type ), tooltip );
	}
	void NamedParameterTreeItem::addParameter( NamedParameterTreeItem* item ) {
		qWarning() << "NamedParameterTreeItem::addParameter(" << item << ")\n * Is NOT working! *";
	}

	Qt::ItemFlags NamedParameterTreeItem::guiFlags() const {
		return QTreeItem::guiFlags() | Qt::ItemIsEditable | Qt::ItemIsSelectable;
	}
	QStringList NamedParameterTreeItem::guiActions() const {
		if ( guiHint() & NamedParameterTreeItem::ProvideReset )
			return QTreeItem::guiActions() << "Reset";
		return QTreeItem::guiActions();
	}
	void NamedParameterTreeItem::guiExecAction( QString action, GuiServer_Interface* gui ) {
		qDebug() << "NamedParameterTreeItem::guiExecAction(" << action << "," << gui <<")";
		if ( action == "Reset" ) {
			setValue( 0 );
		} else
			QTreeItem::guiExecAction( action, gui );
	}
	QVariant NamedParameterTreeItem::guiIdentifier( int role ) const {
		switch ( role ) {
			case Qt::ToolTipRole:
				return _tooltip;
				break;
			default:
				return QTreeItem::guiIdentifier( role );
		}
		return QVariant();
	}

	QString NamedParameterTreeItem::toXML() {
		QString ret = "";
		foreach( QString listitem, childlist() ) {
			NamedParameterTreeItem* tmp = dynamic_cast<NamedParameterTreeItem*>( getNamedChild( listitem ) );
			if ( tmp ) {
				ret += tmp->toXML();
			}
		}
		QString own = QString( "<parameter name=\"%1\" type=\"%2\" " ).arg( objectName() ).arg( QVariant::typeToName( type() ) );
		if ( type() != QVariant::Invalid )
			own += QString( "value=\"%1\" " ).arg( toString() );
		if ( ret != "" )
			ret = own + ">" + ret + "</parameter>";
		else
			ret = own + "/>";
		return ret;
	}

	void NamedParameterTreeItem::setToolTip( QString n ) {
		_tooltip = n;
	}

	void setValueFromNode( NamedParameterTreeItem* item, QDomNode node ) {
		NamedParameterTreeItem::setValueFromNode( item, node );
	}
	void NamedParameterTreeItem::setValueFromNode( NamedParameterTreeItem* item, QDomNode node ) {
		//qDebug() << "setValueFromNode(" << item << "," << node.nodeName() << ")";
		if ( node.nodeName() == "parameter" && item != 0 ) {
			//qDebug() << " Yes, its a parameter!";

			QString name, type, value;

			if ( node.attributes().contains( "name" ) )
				name = node.attributes().namedItem( "name" ).nodeValue();
			if ( node.attributes().contains( "type" ) )
				type = node.attributes().namedItem( "type" ).nodeValue();
			if ( node.attributes().contains( "value" ) )
				value = node.attributes().namedItem( "value" ).nodeValue();

			//qDebug() << " Name is" << name << " type is" << QVariant::nameToType( type.toStdString().c_str() ) << " value is" << value;
			//qDebug() << " Own name is" << item->objectName() << " type is" << item->type() << " own value is" << item->value();

			if ( name == item->objectName() && type == QVariant::typeToName( item->type() ) ) {
				QVariant tmp = value;
				tmp.convert( item->type() );
				item->setValue( tmp );
			}

			QStringList childlist = item->childlist();
			QDomNode child = node.firstChild();
			while ( ! child.isNull() ) {
				if ( child.attributes().contains( "name" ) ) {
					QString name = child.attributes().namedItem( "name" ).nodeValue();
					if ( childlist.contains( name ) )
						setValueFromNode( dynamic_cast<NamedParameterTreeItem*>( item->child( name ) ), child );
				}
				child = child.nextSibling();
			}
		}
	}
};

