/*
    Copyright 2005-2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "abstractdiskio.h"
#include "abstractdiskio.moc"

#include <QtCore/QDebug>
#include <QtCore/QFile>
#include <QtCore/QCoreApplication>
#include <QtCore/QThread>
#include <QtCore/QTime>
#include <QtCore/QTimer>
#include <QtGui/QIcon>

namespace NSDiskIO {

	AbstractReader::AbstractReader( QObject* p ) 
		: NSCore::DataSetProducer( p )
		, _reading( false )
		, _stopReading( false )
		, _time( new QTime( 0, 0 ) )
		, _datasetsToRead( 0 )
		, _readAllWaitForEmpty( false )
	{
	}
	AbstractReader::~AbstractReader() {
	}

	void AbstractReader::readDataSet() {
		readNDataSets( 1 );
	}
	void AbstractReader::readAllDataSets() {
		if ( !_reading ) {
			_reading = true;
			emit startedRead();

			if ( thread() != QThread::currentThread() )
				qWarning() << "AbstractReader::readAllDataSets() NOT running in my own thread.";

			qDebug() << "Starting to read ALL datasets from the file";

			QTimer::singleShot( 10, this, SLOT( privateReadAllDataSets() ) );
		}
	}
	void AbstractReader::privateReadAllDataSets() {
		if ( thread() != QThread::currentThread() )
			qWarning() << "AbstractReader::privateReadAllDataSets() Not running in my thread!!!";
		if ( atEnd() || _stopReading == true ) {
			qDebug() << "Done.";
			_reading = false;
			_stopReading = false;
			emit stoppedRead();
		} else {
			readSingleDataSet();
			QTimer::singleShot( 0, this, SLOT( privateReadAllDataSets() ) );
		}
	}

	void AbstractReader::readNDataSets( int n ) {
		if ( _reading )
			return;
		if ( _datasetsToRead == 0 )
			QTimer::singleShot( 20, this, SLOT( privateReadPendingDataSets() ) );
		_datasetsToRead += n;
	}
	void AbstractReader::privateReadPendingDataSets() {
		if ( _datasetsToRead < 1 )
			return;

		//if ( thread() != QThread::currentThread() )
		//	qWarning() << "AbstractReader::privateReadPendingDataSets() Not running in my thread!!!";

		readSingleDataSet();
		_datasetsToRead--;

		if ( QThread::currentThread() == QCoreApplication::instance()->thread() )
			/// Single-thread application
			QTimer::singleShot( 20, this, SLOT( privateReadPendingDataSets() ) );
		else
			QTimer::singleShot( 0, this, SLOT( privateReadPendingDataSets() ) );

	}
	void AbstractReader::readTenDataSets() {
		readNDataSets( 10 );
	}
	void AbstractReader::readHundredDataSets() {
		readNDataSets( 100 );
	}
	void AbstractReader::signalEmpty() {
		emit queueEmpty();
	}
	void AbstractReader::addToTime( int n ) {
		QTime old( *_time );
		delete _time;
		_time = new QTime( old.addMSecs( n ) );
	}
	QString AbstractReader::toXML() const {
		QString ret = "<reader filename=\"%1\" />";
		return ret.arg( filename() );
	}
	void AbstractReader::fromXML( QString xml ) {
		QDomDocument doc( "abstractreader-doc" );
		doc.setContent( xml );
		fromXML( doc.documentElement() );
	}
	void AbstractReader::fromXML( QDomNode node ) {
		QDomNamedNodeMap items = node.attributes();
		if ( items.contains( "filename" ) )
			setFile( items.namedItem( "filename" ).toAttr().value() );
	}
	QVariant AbstractReader::guiIdentifier( int role ) const {
		switch ( role ) {
			case Qt::DecorationRole:
				return QIcon( ":/icons/cr48-action-load_pulses.png" );
				break;
			default:
				return NSCore::DataSetProducer::guiIdentifier( role );
		}
	}

	AbstractWriter::AbstractWriter( QObject* p )
		: NSCore::DataSetChainMember( p )
		, _number_of_written_datasets( 0 )
	{
	}
	void AbstractWriter::dataAvailable( NSCore::DataSetProducer* producer ) {
		qDebug() << "AbstractWriter::dataAvailable(" << producer << ") [defunct]";
		//writeDataSet( *( producer->first() ) );
		//producer->deleteFirst();
		++_number_of_written_datasets;
		emit wroteDataSet( _number_of_written_datasets );
	}

	void AbstractWriter::resetCounter() {
		_number_of_written_datasets = 0;
	}

	QString AbstractWriter::toXML() const {
		QString ret = "<writer />"; //"<writer filename=\"%1\" />";
		return ret/*.arg( filename() )*/;
	}


}

