/*
    Copyright 2005-2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "evaluationpath.h"
#include "evaluationpath.moc"

#include "evaluationpath_items.h"
#include "evaluationpath_items.moc"
#include "plugins_interface.h"

#include "mathexpression.h"
#include "filterexpression.h"

#include <QtCore/QDebug>
#include <QtCore/QCoreApplication>
#include <QtCore/QThread>
#include <QtCore/QTimer>
#include <QtXml/QDomDocument>
#include <QtGui/QIcon>

namespace NSEvaluation {

EvaluationPath::EvaluationPath( QObject* p, NSCore::QTreeItem* t )
	: NSCore::DataSetChainMember( t )
	, _config( new NSCore::NamedParameterTreeItem( "config", QVariant::Invalid, this ) )
	, _timer( new QTimer( this ) )
{
	qDebug() << "EvaluationPath::EvaluationPath(" << p << ")";

	_config->addParameter( "processed", QVariant( 0 ) );
	_config->getParameter( "processed" )->setMin( 0 );
	_config->getParameter( "processed" )->setMax( int( 1E10 ) );
	new FilterExpression( _config );
	//_config->addParameter( "through", QVariant( 0 ) );
	//_config->getParameter( "through" )->setMin( 0 );
	//_config->getParameter( "through" )->setMax( int( 1E10 ) );
	_config->getParameter( "processed" )->setGuiHint( NSCore::NamedParameterTreeItem::Counter );
	//_config->getParameter( "through" )->setGuiHint( NSCore::NamedParameterTreeItem::Counter );

	qDebug() << "EvaluationPath::EvaluationPath() finished";

	_timer->setInterval( 1000 );
	//_timer->setSingleShot( true );
	connect( _timer, SIGNAL( timeout() ), this, SLOT( checkBuffers() ) );
	_timer->start();
}
EvaluationPath::~EvaluationPath() {
	qDebug() << "EvaluationPath::~EvaluationPath()";
	_timer->stop();
}

bool EvaluationPath::dataArrived( NSCore::AcqisitionDataSet* data ) {
	//qDebug() << "EvaluationPath::dataArrived(" << data << ")";
	_config->getParameter( "processed" )->setValue( _config->getParameter( "processed" )->toInt() + 1 );

	foreach( PluginItem* plugin, childobjects<PluginItem>() )
		plugin->process( *data );

	if ( _config->child<FilterExpression>( "Filter" )->value( *data ) > 0 )
		return true;

	return false;
}

void EvaluationPath::checkBuffers() {
	if ( thread() != QThread::currentThread() )
		qDebug() << "EvaluationPath::checkBuffers() my thread:" << thread() << " current thread:" << QThread::currentThread();

	//_config->child<FilterExpression>( "Filter" )->workOnCache();

	while ( ! _pluginsToDelete.isEmpty() )
		delete _pluginsToDelete.takeFirst();

	while ( ! _xmlToParse.isEmpty() )
		parseXML( _xmlToParse.takeFirst() );

	while ( ! _pluginsToAdd.isEmpty() ) {
		newPlugin( _pluginsToAdd.begin().key(), _pluginsToAdd.begin().value() );
		_pluginsToAdd.erase( _pluginsToAdd.begin() );
	}

	if ( ! _timer->isActive() )
		_timer->start();
}

void EvaluationPath::execAdd( NSCore::QTreeItem* parent, NSCore::GuiServer_Interface* gui ) {
	QString plugin = gui->getItem( "Select the new plugin", "Select the new plugin you want to add to the tree:", PluginFactory::the()->plugins() );
	if ( ! plugin.isEmpty() )
		newPlugin( plugin, parent );
}
void EvaluationPath::execRemove( NSCore::QTreeItem* p, NSCore::GuiServer_Interface* gui ) {
	PluginItem* plugin = dynamic_cast<PluginItem*>( p );
	if ( !plugin )
		return;
	bool yes = gui->messageYesNo( "Are you sure?", "Do you really want to remove this plugin?" );
	if ( yes )
		deletePlugin( plugin );
}

PluginItem* EvaluationPath::newPlugin( QString plugin, QTreeItem* parent ) {
	qDebug() << "EvaluationPath::newPlugin(" << plugin << "," << parent << ")";
	if ( thread() != QThread::currentThread() ) {
		qDebug() << " wrong thread, will cache the action...";
		_pluginsToAdd[ plugin ] = dynamic_cast<PluginItem*>( parent );
		return 0;
	}

	if ( plugin == "" )
		return 0;

	if ( parent == 0 )
		parent = this;

	QString plugintype = PluginFactory::the()->plugintype( plugin );
	PluginItem* ret =0;

	if ( plugintype != "" ) {
		qDebug() << " plugintype =" << plugintype;
		if ( plugintype == "Analyzer" )
			ret = new AnalyzerItem( plugin, parent );
		if ( plugintype == "Modifier" )
			ret = new ModifierItem( plugin, parent );
		if ( plugintype == "Filter" )
			ret = new FilterItem( plugin, parent );
	}

	if ( ret != 0 )
		emit newPlugin( ret, parent );

	return ret;
}

bool EvaluationPath::deletePlugin( PluginItem* plugin ) {
	//qDebug() << "EvaluationPath::deletePlugin(" << plugin << ")";

	// No need to continue with no item to delete.
	if ( plugin == 0 )
		return false;
	// Don't delete yourself, you are the root of the path.
	if ( dynamic_cast<QTreeItem*>( plugin ) == this )
		return false;

	_pluginsToDelete.append( plugin );

	return true;
}
void EvaluationPath::guiExecAction( QString action, NSCore::GuiServer_Interface* gui ) {
	if ( action == "Add" )
		execAdd( this, gui );
	else
		NSCore::QTreeItem::guiExecAction( action, gui );
}
QVariant EvaluationPath::guiIdentifier( int role ) const {
	switch ( role ) {
		case Qt::DisplayRole:
			return "EvaluationPath";
			break;
		case Qt::DecorationRole:
			return QIcon( ":/icons/cr48-action-evaluate_pulses.png" );
			break;
		default:
			return QTreeItem::guiIdentifier( role );
	}
}

void EvaluationPath::parseXML( const QString& xmlstring ) {
	QDomDocument doc( "evaluationpath-document" );
	doc.setContent( xmlstring );
	parseXML( doc.documentElement() );
}

void EvaluationPath::parseXML( QDomNode node ) {
	if ( node.nodeName() == "evaluationpath" ) {
		if ( thread() != QThread::currentThread() ) {
			_xmlToParse.append( node );
			foreach( QObject* obj, childobjects() ) {
				PluginItem* tmp = dynamic_cast<PluginItem*>( obj );
				if ( tmp )
					_pluginsToDelete.append( tmp );
			}
		} else
			internalParseXML( node, this );
	}
}

void EvaluationPath::internalParseXML( QDomNode node, NSCore::QTreeItem* parent ) {
	QString type;
	for ( int i=0; i<node.attributes().size(); ++i ) {
		if ( node.attributes().item( i ).nodeName() == "type" )
			type = node.attributes().item( i ).nodeValue();
	}
	if ( !( node.nodeName() == "analyzer" || node.nodeName() == "filter" || node.nodeName() == "modifier" || node.nodeName() == "evaluationpath" ) ) {
		return;
	}

	PluginItem* plugin = 0;
	plugin = newPlugin( type, parent );
	qWarning() << "EvaluationPath::internalParseXML() plugin =" << plugin;
	if ( node.nodeName() == "filter" || node.nodeName() == "modifier" )
		parent = plugin;

	QDomNode options = node.firstChildElement( "parameter" );
	if ( ! options.isNull() && plugin ) {
		qDebug() << "EvaluationPath::internalParseXML(...) found a <parameter />:";
		NSCore::NamedParameterTreeItem::setValueFromNode( *( plugin->config() ), options );
	}

	QDomNode child = node.firstChild();
	while ( ! child.isNull() ) {
		internalParseXML( child, parent );
		child = child.nextSibling();
	}
}


QString EvaluationPath::toXML() const {
	QString ret;
	for ( int i=0; i<childs(); ++i ) {
		PluginItem* p = dynamic_cast<PluginItem*>( child( i ) );
		if ( p )
			ret += p->toXML();
	}
	if ( ! ret.isNull() )
		return QString( "<evaluationpath>%1</evaluationpath>" ).arg( ret );
	return "<evaluationpath />";
}

};

