/*
    Copyright 2005-2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "xyspecview.h"
#include "xyspecview.moc"

#include <cmath>
#include <QtCore/QDebug>
#include <QtGui/QPainter>
#include <QtGui/QHBoxLayout>
#include <QtGui/QLabel>

#ifdef HAVE_QWT
#include <qwt5/qwt_plot.h>
#include <qwt5/qwt_plot_curve.h>
#include <qwt5/qwt_plot_spectrogram.h>

double abs( const QPointF& n ) {
	return sqrt( pow( n.x(),2 ) + pow( n.y(),2 ) );
}
bool operator<( const QPointF& x, const QPointF& y ) {
	return ( abs( x ) < abs( y ) );
}

QDebug operator<<( QDebug stream, QwtDoubleInterval& interval ) {
	stream << "QwtDoubleInterval(" << interval.minValue() << "," << interval.width() << ")";
	return stream;
}

class SpecData: public QwtRasterData {
	public:
		SpecData() : model( 0 ), zrange( new QwtDoubleInterval( 0, 1 ) ) {
			//qDebug() << "SpecData::SpecData()";
			xbounds.setInterval( 0,0 );
			ybounds.setInterval( 0,0 );
			xsteps = ysteps = 0;
		}
		~SpecData() {
			//qDebug() << "SpecData::~SpecData()";
		}
		QwtRasterData* copy() const {
			//qDebug() << "SpecData::copy()";
			return new SpecData( *this );
		}
		double value( double x, double y ) const {
			//qDebug() << "SpecData::value(" << x << "," << y << ")";
			double xstep = xbounds.width() / xsteps;
			double ystep = ybounds.width() /ysteps;
			int _x = ( x - xbounds.minValue() ) / xstep;
			int _y = ( y - ybounds.minValue() ) / ystep;
			//qDebug() << " value at (" << _x << "," << _y << ") is ";
			QVariant tmp = model->data( model->index( _x, _y ), Qt::DisplayRole );
			//if ( tmp.toDouble() != 0.0 )
			//	qDebug() << "SpecData::value(" << x << "," << y << ") [" << _x << _y << "] returning" << tmp.toDouble();
			return tmp.toDouble();
		}
		QwtDoubleInterval range() const {
			//qDebug() << QString( "SpecData::range() returning [%1,%2]" ).arg( zrange->minValue() ).arg( zrange->maxValue() );
			return *zrange;
		}
		QSize rasterHint( const QwtDoubleRect& ) const {
			return QSize( xsteps, ysteps );
		}

		void extendZRange( double z ) {
			if ( ! zrange->contains( z ) )
				zrange->setMaxValue( z );
		}

		QAbstractItemModel* model;
		QwtDoubleInterval xbounds, ybounds;
		int xsteps, ysteps;
		QwtDoubleInterval* zrange;

		void updateCache() {
			qDebug() << "SpecData::updateCache()";
			xbounds.setInterval( model->headerData( 0, Qt::Horizontal ).toDouble(), model->headerData( model->columnCount()-1, Qt::Horizontal ).toDouble() );
			ybounds.setInterval( model->headerData( 0, Qt::Vertical ).toDouble(), model->headerData( model->rowCount()-1, Qt::Vertical ).toDouble() );
			setBoundingRect( QwtDoubleRect( xbounds.minValue(), ybounds.minValue(), xbounds.width(), ybounds.width() ) );
			xsteps = model->columnCount();
			ysteps = model->rowCount();
			/*qDebug() << " xbounds is now" << xbounds;
			qDebug() << " ybounds is now" << ybounds;
			qDebug() << " boundRect is now" << boundingRect();*/
		}
};
#else
	class SpecData;
	class QwtPlot;
	class QwtPlotSpectrogram;
#endif

class XYSpecView_Private {
	public:
		QwtPlot* plot;
		//QwtPlotCurve* curve;
		QwtPlotSpectrogram* spec;
		SpecData* specdata;
		QHBoxLayout* layout;
};

XYSpecView::XYSpecView( QWidget* p ) : QAbstractItemView( p ), d( new XYSpecView_Private() ) {
	qDebug() << "XYSpecView::XYSpecView(" << p << ")";
	d->layout = new QHBoxLayout( this->viewport() );
#ifdef HAVE_QWT
	d->plot = new QwtPlot( this->viewport() );
	d->layout->addWidget( d->plot );
	/*double x[ 10 ], y[ 10 ];
	for ( int i=0; i<10; ++i ) {
		x[ i ] = i/10.0;
		y[ i ] = i*i/100.0;
	}
	d->curve = new QwtPlotCurve( "Curve" );

	d->curve->setData( x, y, 10 );
	d->curve->attach( d->plot );*/

	d->spec = new QwtPlotSpectrogram( "Spec" );
	d->spec->attach( d->plot );

	d->specdata = new SpecData();
	d->specdata->setBoundingRect( QRectF( -1, -1, 2, 2 ) );
	d->spec->setData( *( d->specdata ) );
#else
	QLabel* lbl = new QLabel( "<qt><center>Sorry, without <i>qwt</i> I can't display the spectrum.</center></qt>", this );
	d->layout->addWidget( lbl );
#endif
}
XYSpecView::~XYSpecView() {
	qDebug() << "XYSpecView::~XYSpecView()";
}

QRect XYSpecView::visualRect( const QModelIndex& ) const {
	return QRect();
}
void XYSpecView::scrollTo( const QModelIndex&, ScrollHint ) {
}
QModelIndex XYSpecView::indexAt( const QPoint& ) const {
	return QModelIndex();
}
void XYSpecView::setModel( QAbstractItemModel* model ) {
	qDebug() << "XYSpecView::setModel(" << model << ")";
	d->specdata->model = model;
	d->specdata->updateCache();
	QAbstractItemView::setModel( model );
	d->spec->setData( *( d->specdata ) );
	viewport()->update();
}

QModelIndex XYSpecView::moveCursor( QAbstractItemView::CursorAction, Qt::KeyboardModifiers ) {
	return QModelIndex();
}

int XYSpecView::horizontalOffset() const {
	return 0;
}
int XYSpecView::verticalOffset() const {
	return 0;
}

bool XYSpecView::isIndexHidden( const QModelIndex& ) const {
	return false;
}

void XYSpecView::setSelection( const QRect&, QItemSelectionModel::SelectionFlags ) {
}

QRegion XYSpecView::visualRegionForSelection( const QItemSelection& ) const {
	return QRegion();
}

void XYSpecView::dataChanged( const QModelIndex& _topLeft, const QModelIndex& _bottomRight ) {
	//qDebug() << "XYSpecView::dataChanged(" << _topLeft << "," << _bottomRight << ")";
#ifdef HAVE_QWT

	for ( int x=_topLeft.column(); x<=_bottomRight.column(); ++x )
		for ( int y=_topLeft.row(); y<=_bottomRight.row(); ++y )
			d->specdata->extendZRange( model()->data( model()->index( x,y ) ).toDouble() );
	//d->spec->setData( *( d->specdata ) );
	//d->plot->replot();
	viewport()->update();
#endif
}
void XYSpecView::paintEvent( QPaintEvent* ev ) {
	d->plot->replot();
}

void XYSpecView::rowsAboutToBeRemoved( const QModelIndex& index, int start, int stop ) {
	qDebug() << "XYSpecView::rowsAboutToBeRemoved(" << index << "," << start << "," << stop << ")";
}
void XYSpecView::rowsInserted( const QModelIndex& index, int start, int stop ) {
	qDebug() << "XYSpecView::rowsInserted(" << index << "," << start << "," << stop << ")";
	d->specdata->updateCache();
	viewport()->update();
}

