/*
    Copyright 2005-2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "modifiers.h"
//#include "modifiers.moc"

#include <math.h>
#include <QtCore/QDebug>
#include <analyzers.h>

AbsModifier::AbsModifier( NSCore::QTreeItem* p ) : NSEvaluation::Modifier( p ) {
}

void AbsModifier::modify( NSCore::AcqisitionDataSet& data ) {
	//qDebug() << "AbsModifier::modify( data )";
	//qDebug() << " generations        =" << data.channel( 0 )->generations();
	//qDebug() << " current generation =" << data.channel( 0 )->currentGeneration();

	for ( int j=0; j<data.channels(); ++j ) {
		for ( int i=0; i<data.channel( j )->samples(); ++i ) {
			data.channel( j )->setSample( i, fabs( data.channel( j )->sample( i, -1 ) ) );
			//qDebug() << j << " " << i << ":" << data.channel( j )->sample( i ) << " " << data.channel( j )->sample( i, 0 );
		}
		data.channel( j )->addResult( "modifier/action", "Absolute of the values" );
	}
}

NormalizeModifier::NormalizeModifier( NSCore::QTreeItem* p ) : NSEvaluation::Modifier( p ) {
	for ( int i=0; i<2; ++i ) {
		_config->addParameter( QString( "Channel %1/Height to Normalize" ).arg( i ), 1.0 );
		_config[ QString( "Channel %1/Height to Normalize" ).arg( i ) ]->setMin( -100.0 );
		_config[ QString( "Channel %1/Height to Normalize" ).arg( i ) ]->setMax( 100.0 );
		_config->addParameter( QString( "Channel %1/Offset after Normalize" ).arg( i ), 0.0 );
		_config[ QString( "Channel %1/Offset after Normalize" ).arg( i ) ]->setMin( -100.0 );
		_config[ QString( "Channel %1/Offset after Normalize" ).arg( i ) ]->setMax( 100.0 );
	}
}

void NormalizeModifier::modify( NSCore::AcqisitionDataSet& data ) {
	for ( int i=0; i<data.channels(); ++i ) {

		double height_to_normalize = _config[ QString( "Channel %1/Height to Normalize" ).arg( i%2 ) ]->toDouble();
		double offset_after_normalize = _config[ QString( "Channel %1/Offset after Normalize" ).arg( i%2 ) ]->toDouble();

		double factor = 1/height_to_normalize;

		for ( int j=0; j<data.channel( i )->samples(); ++j )
			data.channel( i )->setSample( j, data.channel( i )->sample( j, -1 ) * factor + offset_after_normalize );

		data.channel( i )->addResult( "modifier/action", QString( "Normalized by factor of %1" ).arg( factor ) );
		data.channel( i )->addResult( "modifier/action", QString( "Added Offset of %1" ).arg( offset_after_normalize ) );

	}
}

FloatingAverageModifier::FloatingAverageModifier( NSCore::QTreeItem* p ) : NSEvaluation::Modifier( p ) {
	qDebug() << "FloatingAverageModifier::FloatingAverageModifier()";
	_config->addParameter( "neighbours", 1 );
	_config[ "neighbours" ]->setMin( 0 );
	_config[ "neighbours" ]->setMax( 10 );
	_config[ "neighbours" ]->setToolTip( "The number of neighbours o the left and right that is looked at." );
}

void FloatingAverageModifier::modify( NSCore::AcqisitionDataSet& data ) {
	int neighbours = _config[ "neighbours" ]->toInt();
	double value;
	for ( int i=0; i<data.channels(); ++i ) {

		for ( int j=0; j<data.channel( i )->samples(); ++j ) {

			value=0;
			for ( int k=-neighbours; k<=neighbours; ++k ) {
				value += data.channel( i )->sample( j+k, -1 );
			}
			value /= 2*neighbours + 1;

			data.channel( i )->setSample( j, value );

		}

	}
}

IntegrateModifier::IntegrateModifier( NSCore::QTreeItem* p ) : NSEvaluation::Modifier( p ) {
}

void IntegrateModifier::modify( NSCore::AcqisitionDataSet& data ) {
	for ( int i=0; i<data.channels(); ++i ) {

		double old = 0;

		for ( int j=0; j<data.channel( i )->samples(); ++j ) {
			//old = data.channel( i )->sample( j, 0 ) + old;
			data.channel( i )->setSample( j, old = data.channel( i )->sample( j, -1 ) + old );
		}
	}
}

PulseIntegrate::PulseIntegrate( NSCore::QTreeItem* p )
	: NSEvaluation::Modifier( p )
	, _analyzer( new InterpolatedMinimumAnalyzer( _config.object() ) )
{
}
void PulseIntegrate::modify( NSCore::AcqisitionDataSet& data ) {
	int before = 20;
	int after = 10;
	for ( int i=0; i<data.channels(); ++i ) {
		if ( data.channel( i )->result( "interpolated minimum at", 0 ) == 0 ) {
			// TODO go back one generation, analyze and return to the current generation
			//_analyzer->analyze( data );
		} else {
			double min_at = data.channel( i )->result( "interpolated minimum at", -1 )->toDouble();
			double old = 0;
			for ( int j = int( min_at - before ); j < min_at + after && j > 0 && j < data.channel( i )->samples(); ++j )
				data.channel( i )->setSample( j, old = data.channel( i )->sample( j, -1 ) + old );
			for ( int j = int( min_at + after ); j < data.channel( i )->samples(); ++j )
				data.channel( i )->setSample( j, old = old - data.channel( i )->sample( j, -1 ) );
		}
	}
	//data.debugPrint();
}

SignedIntegrate::SignedIntegrate( NSCore::QTreeItem* p )
	:NSEvaluation::Modifier( p )
{
}

void SignedIntegrate::modify( NSCore::AcqisitionDataSet& data ) {
	for ( int i=0; i<data.channels(); ++i ) {
		double last = 0; //data.channel( i )->sample( 0, 0 );
		double now = 0, sum = 0;
		//data.channel( i )->setSample( 0, sum );
		for ( int j=0; j<data.channel( i )->samples(); ++j ) {

			now = data.channel( i )->sample( j, -1 );

			if ( now < last )
				sum -= fabs( now );
			else
				sum += fabs( now );

			data.channel( i )->setSample( j, sum );

			last = now;
		}
	}
}

CrossCorrelation::CrossCorrelation( NSCore::QTreeItem* p ) : NSEvaluation::Modifier( p ) {
	_config->addParameter( "middle percentage", 100 );
	_config[ "middle percentage" ]->setMin( 0 );
	_config[ "middle percentage" ]->setMax( 100 );
	_config[ "middle percentage" ]->setToolTip( "The percentage of data that is to be evaluated. Will be aligned in the middle of the dataset..." );
}

void CrossCorrelation::modify( NSCore::AcqisitionDataSet& data ) {
	if ( data.channels() < 2 ) {
		qDebug() << "Dataset doesn't contain enough channels.";
		return;
	}

	double percent = _config->getParameter( "middle percentage" )->toInt() / 100.0;

	double firstamplitude = 1;
	double secondamplitude = 1;
	QVariant* tmp;
	if ( ( tmp = data.channel( 0 )->result( "interpolated minimum", -1 ) )->isValid() )
		firstamplitude = tmp->toDouble();
	if ( ( tmp = data.channel( 1 )->result( "interpolated minimum", -1 ) )->isValid() )
		secondamplitude = tmp->toDouble();

	int ibounds = data.channel( 0 )->samples() - int( data.channel( 0 )->samples() * percent );
	ibounds /= 2;

	for ( int i=0; i<ibounds; ++i )
		data.channel( 0 )->setSample( i, 2 );

	for ( int i=ibounds; i< data.channel( 0 )->samples() - ibounds; ++i ) {

		int joffset = qMax( data.channel( 1 )->samples() / 2 - i, 0 );
		int ioffset = qMax( i - data.channel( 1 )->samples()/2, 0 );
		int jend = data.channel( 1 )->samples()/2 + i;
		if ( jend > data.channel( 0 )->samples() )
			jend = data.channel( 1 )->samples()/2 + data.channel( 0 )->samples() - i;

		double sum = 0;
		double tmp;
		for ( int j=0; j<jend; ++j ) {
			tmp = data.channel( 0 )->sample( ioffset + j, -1 )/firstamplitude - data.channel( 1 )->sample( joffset + j, -1 )/secondamplitude;
			sum += tmp*tmp;
		}

		data.channel( 0 )->setSample( i, sqrt( sum / jend ) );
	}

	for ( int i=data.channel( 0 )->samples() - ibounds; i<data.channel( 0 )->samples(); ++i )
		data.channel( 0 )->setSample( i, 2 );

}

