/*
    Copyright 2005-2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "acqiris.h"
#include "acqiris.moc"
#include <QtCore/QTimer>
#include <QtCore/QDateTime>

#ifdef HAVE_ACQIRIS
#include "AcqirisD1Import.h"
#endif

#include <cmath>

namespace NSAcqiris {

	Acqiris::Acqiris( QObject* p ) : NSCore::QTreeItem( "acqiris", p ) {
		qDebug( "Acqiris::Acqiris()" );
#ifdef HAVE_ACQIRIS
		qDebug( " Acqiris-libs and -headers seemed to be found. This class will be more than a dummy." );
		long nbrInstruments=0;
		ViStatus status;
		ViString options = "";
		status = AcqrsD1_multiInstrAutoDefine( options, &nbrInstruments );
		qDebug( "found instruments: %li - creating the digitizers", nbrInstruments );
		for ( long i=0; i<nbrInstruments; i++ ) {
			ViSession instrumentID=0;
			char name[ 20 ];
			sprintf( name, "PCI::INSTR%li", i );
			status = AcqrsD1_InitWithOptions( name, VI_FALSE, VI_FALSE, options, &instrumentID );
			_digitizers.push_back( new Digitizer( instrumentID, this, QString( "digitizer %1" ).arg( i ) ) );
		}
#endif
		qDebug( "initialized digitizers: %i", int( _digitizers.size() ) );
		debugPrint();
	}

	Acqiris::~Acqiris() {
		qDebug( "Acqiris::~Acqiris()" );
		while ( _digitizers.size()>0 ) {
			delete *( _digitizers.begin() );
			_digitizers.pop_front();
		}
#ifdef HAVE_ACQIRIS
			AcqrsD1_closeAll();
#endif
		qDebug( "Acqiris::~Acqiris() done." );
	}

	Digitizer* Acqiris::digitizer( int n ) {
		std::list<Digitizer*>::iterator it = _digitizers.begin();
		for ( int i=0; i<int( _digitizers.size() ) && it!=_digitizers.end(); ++i ) {
			if ( n == i )
				return ( *it );
			++it;
		}
		return 0;
	}




	Digitizer::Digitizer( ViSession instID, Acqiris* p, QString name )
		: NSCore::DataSetProducer( p )
		, _instrumentID( instID )
		, _c( new NSCore::NamedParameterTreeItem( "config", QVariant::Invalid, this ) )
		, _singleShot( false )
		, _stopped( false )
		, _horizontalConfigNeeded( true )
		, _triggerConfigNeeded( true )
		, _verticalConfigNeeded( true )
		, _waveFormArraySize( 0 )
	{
		setObjectName( name );

		qDebug( "\tDigitizer::Digitizer( ViSession %i )", int( instID ) );

		_c->addParameter( "timeout", 2000 );
		_c[ "timeout" ]->setRange( 1, 3600000 );
		_c->addParameter( "forcetrigger", QVariant( false ) );
		_c->addParameter( "horizontal", QVariant::Invalid );
		_c->addParameter( "horizontal/samples", 100 );
		_c[ "horizontal/samples" ]->setRange( 16, int( pow( 2, 16 ) ) );
		_c->addParameter( "horizontal/samplinginterval", 1/4.0, "Sampling interval in units of picoseconds." );
		_c[ "horizontal/samplinginterval" ]->setRange( 1/8.0, 1000 );
		_c->addParameter( "horizontal/segments", 1 );
		_c[ "horizontal/segments" ]->setRange( 1, 1024 );
		_c->addParameter( "horizontal/delaytime", 0.0, "The trigger-delay in units of picoseconds." );
		_c[ "horizontal/delaytime" ]->setRange( -100.0, 100.0 );
		connect( _c[ "horizontal" ], SIGNAL( valueChanged() ), this, SLOT( scheduleHorizontalConfig() ) );

		_c->addParameter( "trigger", QVariant::Invalid );
		_c->addParameter( "trigger/channel", 1,
			"Positive numbers for the inputchannels, negative numbers for the trigger-inputs." );
		_c[ "trigger/channel" ]->setRange( -4, 2 );
		_c->addParameter( "trigger/slope", QVariant::Int,
			"0: Positive <br>1: Negative <br>2: Out of Window <br>3: Into Window <br>4: HF Divide" );
		_c[ "trigger/slope" ]->setRange( 0, 4 );
		_c->addParameter( "trigger/level", 0.0 );
		//_c[ "trigger/level" ]->setRange( -100.0, 100.0 );
		//_c[ "trigger/level" ]->setGuiHint( NSCore::Parameter::ReadAllowed );
		_c->addParameter( "trigger/levelintern", 0.0,
			"The trigger-level for internal triggermode in percent." );
		_c[ "trigger/levelintern" ]->setRange( -100.0, 100.0 );
		connect( _c[ "trigger/levelintern" ], SIGNAL( valueChanged( double ) ),
		         this, SLOT( triggerLevelIntern( double ) ) );
		_c->addParameter( "trigger/levelextern", 0.0,
			"The trigger-level for external triggers in Volts." );
		_c[ "trigger/levelextern" ]->setRange( -5.0, 5.0 );
		connect( _c[ "trigger/levelextern" ], SIGNAL( valueChanged( double ) ),
		         this, SLOT( triggerLevelExtern( double ) ) );
		_c->addParameter( "trigger/level2", 0.0 );
		_c[ "trigger/level2" ]->setRange( -100.0, 100.0 );
		_c->addParameter( "trigger/coupling", QVariant::Int,
			"0: DC <br>1: AC <br>2: HF Reject <br>3: DC 50Ohm <br>4: AC 50Ohm" );
		_c[ "trigger/coupling" ]->setRange( 0, 4 );
		connect( _c[ "trigger" ], SIGNAL( valueChanged() ), this, SLOT( scheduleTriggerConfig() ) );

		for ( int i=0; i<channels(); i++ ) {
			qDebug() << " Adding Channel" << i+1;
			_c->addParameter( QString( "Channel %1" ).arg( i+1 ), QVariant::Invalid );
			_c->addParameter( QString( "Channel %1/fullscale" ).arg( i+1 ), "5V" );
			_c->addParameter( QString( "Channel %1/offset" ).arg( i+1 ), 0.0 );
			_c[ QString( "Channel %1/offset" ).arg( i+1 ) ]->setRange( -5.0, 5.0 );
			_c->addParameter( QString( "Channel %1/coupling" ).arg( i+1 ), 3 );
			_c->addParameter( QString( "Channel %1/bandwidth" ).arg( i+1 ), QVariant( 0 ) );
			connect( _c[ QString( "Channel %1" ).arg( i+1 ) ], SIGNAL( valueChanged() ), this, SLOT( scheduleVerticalConfig() ) );
		}
		for ( int i=0; i<externaltriggers(); ++i ) {
			qDebug() << " Adding Channel" << -i-1;
			_c->addParameter( QString( "Channel %1" ).arg( -i-1 ), QVariant::Invalid );
			_c->addParameter( QString( "Channel %1/fullscale" ).arg( -i-1 ), "5V" );
			_c->addParameter( QString( "Channel %1/offset" ).arg( -i-1 ), 0.0 );
			_c[ QString( "Channel %1/offset" ).arg( -i-1 ) ]->setRange( -5.0, 5.0 );
			connect( _c[ QString( "Channel %1" ).arg( -i-1 ) ], SIGNAL( valueChanged() ), this, SLOT( scheduleVerticalConfig() ) );
		}

		_c->addParameter( "LED color", "Off" );
		connect( _c[ "LED color" ], SIGNAL( valueChanged( QString ) ), this, SLOT( setLEDColor( QString ) ) );

		//_leds.push_back( new LED( _instrumentID, this ) );

#ifdef HAVE_ACQIRIS
		ViInt32 ctrlIOPattern;
		AcqrsD1_getControlIO( _instrumentID, 0, &ctrlIOPattern, NULL, NULL );
#endif
	}

	Digitizer::~Digitizer() {
		qDebug( "\tDigitizer::~Digitizer()" );
#ifdef HAVE_ACQIRIS
		AcqrsD1_stopAcquisition( _instrumentID );
#endif
		while ( !_waveFormArrays.isEmpty() )
			delete _waveFormArrays.takeLast();
		//delete _config;
	}

	QString Digitizer::debug() const {
		return QString( "digitizer %1" ).arg( int( _instrumentID ) );
	}

	void Digitizer::singleShot( bool n ) { _singleShot = n; }

	int Digitizer::channels() {
		static ViInt32 channels = -1;
		if ( channels == -1 ) {
#ifdef HAVE_ACQIRIS
			AcqrsD1_getNbrChannels( _instrumentID, &channels );
#else
			channels = 0;
#endif
		}
		return channels;
	}

	int Digitizer::externaltriggers() {
		static ViInt32 externaltriggers = -1;
		if ( externaltriggers == -1 ) {
#ifdef HAVE_ACQIRIS
			AcqrsD1_getInstrumentInfo( _instrumentID, "NbrExternalTriggers", &externaltriggers );
#else
			externaltriggers = 0;
#endif
		}
		return externaltriggers;
	}

	void Digitizer::startAcqisition() {
//		qDebug( "\tDigitizer::startAcqisition()" );
#ifdef HAVE_ACQIRIS
		ViBoolean done = FALSE;
		/*ViStatus status =*/ AcqrsD1_acqDone( _instrumentID, &done );
		if ( done == VI_TRUE ) {
			configHorizontal();
			configTrigger();
			configVertical();
			AcqrsD1_acquire( _instrumentID );
			_stopped = false;
			QTimer::singleShot( 20, this, SLOT( checkForData() ) );
			emit acquisitionStarted();
			_time_since_last_start.start();
		}
		else
			qWarning( "\nDigitizer::startAcqisition() : already an acquisition running!\n" );
#endif
	}

	void Digitizer::stopAcqisition() { _stopped = true; }

	void Digitizer::calibrate() {
		stopAcqisition();
#ifdef HAVE_ACQIRIS
		AcqrsD1_calibrate( _instrumentID );
#endif
		emit acquisitionStopped();
	}

	void Digitizer::checkForData() {
#ifdef HAVE_ACQIRIS
		if ( !_stopped ) {
//			qDebug( "\tDigitizer::checkForData()" );
			ViBoolean done;
			AcqrsD1_acqDone( _instrumentID, &done );
			if ( done == VI_TRUE ) {

				long segments, samples_in_segment, currentPad;
				AcqrsD1_getMemory( _instrumentID, &samples_in_segment, &segments );
				AcqrsD1_getInstrumentInfo( _instrumentID, "TbSegmentPad", &currentPad );
				currentPad *= 4;

				//qDebug() << "\tDigitizer::checkForData() - having" << segments << "segments to read";

				if ( ( samples_in_segment+currentPad )*( segments+1 ) != _waveFormArraySize ) {
					while ( !_waveFormArrays.isEmpty() )
						delete _waveFormArrays.takeLast();
					qDebug( "\tDigitizer::checkForData() need to resize array!" );
				}
				while ( _waveFormArrays.size() < channels() ) {
					_waveFormArrays.push_back( new ViReal64[ ( samples_in_segment+currentPad )*( segments+1 ) ] );
					_waveFormArraySize = ( samples_in_segment+currentPad )*( segments+1 );
				}

				static long _old_segments = -1;
				static double* horPos = 0;
				static long* timeLo = 0;
				static long* timeHi = 0;
				if ( _old_segments != segments ) {
					delete horPos;
					delete timeHi;
					delete timeLo;
					horPos = new double[ segments ];
					timeLo = new long[ segments ];
					timeHi = new long[ segments ];
					_old_segments = segments;
				}
				static int _old_channels = -1;
				static double* vFullScale = 0;
				static double* vOffset = 0;
				if ( _old_channels != channels() ) {
					_old_channels = channels();
					delete vFullScale;
					delete vOffset;
					vFullScale = new double[ _old_channels ];
					vOffset = new double[ _old_channels ];
				}
				static double sampTime;
				static ViInt32 _vBandwidth_tmp, _vCoupling_tmp;

				ViStatus status( 0 );
				ViChar msg[ 500 ];
				for ( int i=0; i<channels(); ++i ) {
					AcqrsD1_getVertical( _instrumentID, i+1, & vFullScale[ i ], & vOffset[ i ], &_vCoupling_tmp, &_vBandwidth_tmp );
					status = AcqrsD1_readRealSequence( _instrumentID, i+1, 0, segments, 0, samples_in_segment, samples_in_segment + currentPad, _waveFormArraySize, _waveFormArrays[ i ], horPos, &sampTime, timeLo, timeHi );
					if ( status != 0 ) {
						AcqrsD1_errorMessage( _instrumentID, status, msg );
						qDebug( "AcqrsD1_readRealSequence - Status: %s", msg );
					}
				}

				if ( !_singleShot )
					startAcqisition();
				else {
					_stopped = true;
					emit acquisitionStopped();
				}

				for ( int i=0; i<segments; ++i ) {
					NSCore::AcqisitionDataSet* data = new NSCore::AcqisitionDataSet( channels(), samples_in_segment );
					data->parameter( "samplinginterval" )->setValue( sampTime );
					data->parameter( "triggerdelay" )->setValue( _c[ "horizontal/delaytime" ]->toDouble() / sampTime * 1E-9 );
					data->parameter( "horizontalposition" )->setValue( horPos[ i ] /sampTime );
					for ( int j=0; j<channels(); ++j ) {
						NSCore::ChannelDataSet* channel = data->channel( j );
						channel->parameter( "fullscale" )->setValue( vFullScale[ j ] );
						channel->parameter( "offset" )->setValue( vOffset[ j ] );
						memcpy( channel->sampledata(), &( _waveFormArrays[ j ][ i*( samples_in_segment+currentPad ) ] ), sizeof( ViReal64 )* samples_in_segment );
						int max_at = 0;
						double max = -1000;
						int min_at = 0;
						double min = 1000;
						for ( int c=0; c<samples_in_segment; ++c ) {
							if ( _waveFormArrays[ j ][ i*(  samples_in_segment+currentPad ) + c ] > max ) {
								max = _waveFormArrays[ j ][ i*(  samples_in_segment+currentPad ) + c ];
								max_at = c;
							}
							if ( _waveFormArrays[ j ][ i*(  samples_in_segment+currentPad ) + c ] < min ) {
								min = _waveFormArrays[ j ][ i*(  samples_in_segment+currentPad ) + c ];
								min_at = c;
							}
						}
						channel->results()->getQVariant( "generation 0/max" )->setValue( max );
						channel->results()->getQVariant( "generation 0/max at" )->setValue( max_at );
						channel->results()->getQVariant( "generation 0/min" )->setValue( min );
						channel->results()->getQVariant( "generation 0/min at" )->setValue( min_at );
					}
					enqueue( data );
				}

			} else {

				if ( _time_since_last_start.elapsed() < _c[ "timeout" ]->toInt() ) 
					QTimer::singleShot( 20, this, SLOT( checkForData() ) );
				else {

					if ( _c[ "forcetrigger" ]->toBool() ) {
						AcqrsD1_forceTrig( _instrumentID );
						QTimer::singleShot( 20, this, SLOT( checkForData() ) );
					} else {
						AcqrsD1_stopAcquisition( _instrumentID );
						qWarning( "Acquisition timed out!" );
						emit acquisitionStoppedTimeout();
					}

				}
				
			}
		} else {
			_stopped = false;
			AcqrsD1_stopAcquisition( _instrumentID );
			emit acquisitionStopped();
		}
#endif
	}

	void Digitizer::configHorizontal() {
		if ( _horizontalConfigNeeded ) {
			qDebug( "\tDigitizer::configHorizontal() need to reconfigure" );
#ifdef HAVE_ACQIRIS
			ViChar msg[ 500 ];
			ViStatus status = AcqrsD1_configHorizontal( _instrumentID,
					_c[ "horizontal/samplinginterval" ]->value().toDouble() * 1E-9,
					_c[ "horizontal/delaytime" ]->value().toDouble() * 1E-9 );
			if ( status != 0 ) {
				AcqrsD1_errorMessage( _instrumentID, status, msg );
				qDebug( "\tAcqrsD1_configHorizontal - Status: %s", msg );
			}
			status = AcqrsD1_configMemory( _instrumentID,
					_c[ "horizontal/samples" ]->value().toInt(),
					_c[ "horizontal/segments" ]->value().toInt() );
			if ( status != 0 ) {
				AcqrsD1_errorMessage( _instrumentID, status, msg );
				qDebug( "\tAcqrsD1_configMemory - Status: %s", msg );
			}
#endif
			_horizontalConfigNeeded = false;
		}
	}

	void Digitizer::configTrigger() {
		if ( _triggerConfigNeeded ) {
			qDebug( "\tDigitizer::configTrigger() need to reconfigure" );
#ifdef HAVE_ACQIRIS
			ViStatus status;
			ViChar msg[ 500 ];
			status = AcqrsD1_configTrigClass( _instrumentID,
					0,
					triggerSourceToPattern( _c[ "trigger/channel" ]->value().toInt() ),
					0,0,0,0 );
			if ( status != 0 ) {
				AcqrsD1_errorMessage( _instrumentID, status, msg );
				qDebug( "\tAcqrsD1_configTrigClass - Status: %s", msg );
			}
			_c[ "trigger/level" ]->setValue( _c[ "trigger/levelintern" ]->value() );
			if ( _c[ "trigger/channel" ]->toInt() < 0 ) {
				qDebug() << " External trigger. level will be" << _c[  "trigger/levelextern" ]->value().toDouble()*1000;
				_c[ "trigger/level" ]->setValue( _c[ "trigger/levelextern" ]->value().toDouble()*1000 );
			}
			status = AcqrsD1_configTrigSource( _instrumentID,
					_c[ "trigger/channel" ]->value().toInt(),
					_c[ "trigger/coupling" ]->value().toInt(),
					_c[ "trigger/slope" ]->value().toInt(),
					_c[ "trigger/level" ]->value().toDouble(),
					_c[ "trigger/level2" ]->value().toDouble() );
			if ( status != 0 ) {
				AcqrsD1_errorMessage( _instrumentID, status, msg );
				qDebug( "\tAcqrsD1_configTrigSource - Status: %s", msg );
			}
#endif
			_triggerConfigNeeded = false;
		}
	}

	void Digitizer::configVertical() {
		if ( _verticalConfigNeeded ) {
#ifdef HAVE_ACQIRIS
			for ( int i=1; i<=channels(); ++i ) {
				qDebug() << "Configuring Channel" << i;
				ViStatus status = AcqrsD1_configVertical( _instrumentID, i,
						fullscale( _c[ QString( "Channel %1/fullscale" ).arg( i ) ]->toString() ),
						_c[ QString( "Channel %1/offset" ).arg( i ) ]->toDouble(),
						_c[ QString( "Channel %1/coupling" ).arg( i ) ]->toInt(),
						_c[ QString( "Channel %1/bandwidth" ).arg( i ) ]->toInt() );
				if ( status != 0 ) {
					ViChar msg[ 500 ];
					//qDebug() << "fullscale =" << _c[ QString( "Channel %1/fullscale" ).arg( i ) ]->toString();
					//qDebug() << "fullscale =" << fullscale( _c[ QString( "Channel %1/fullscale" ).arg( i ) ]->toString() );
					AcqrsD1_errorMessage( _instrumentID, status, msg );
					qDebug( "\tAcqrsD1_configVertical - Channel %i Status: %s", i, msg );
				}
				ViReal64 fullscale, offset;
				ViInt32 coupling,bandwidth;
				status = AcqrsD1_getVertical( _instrumentID, i, &fullscale, &offset, &coupling, &bandwidth );
				if ( status != 0 ) {
					ViChar msg[ 500 ];
					AcqrsD1_errorMessage( _instrumentID, status, msg );
					qDebug( "\tAcqrsD1_getVertical - Channel %i Status: %s", i, msg );
				}
				qDebug() << "\tChannel" << i
					<< endl << "\t\tfullscale =" << fullscale
					<< endl << "\t\toffset =" << offset
					<< endl << "\t\tcoupling =" << coupling
					<< endl << "\t\tbandwidth =" << bandwidth
					<< endl;
			}
			for ( int i=-1; i >= -externaltriggers(); --i ) {
				qDebug() << "Configuring Channel" << i;
				ViStatus status = AcqrsD1_configVertical( _instrumentID, i,
						fullscale( _c[ QString( "Channel %1/fullscale" ).arg( i ) ]->toString() ),
						_c[ QString( "Channel %1/offset" ).arg( i ) ]->toDouble(),
						3,0 );
						/*_c[ QString( "Channel %1/coupling" ).arg( i ) ]->toInt(),
						_c[ QString( "Channel %1/bandwidth" ).arg( i ) ]->toInt() );*/
				if ( status != 0 ) {
					ViChar msg[ 500 ];
					//qDebug() << "fullscale =" << _c[ QString( "Channel %1/fullscale" ).arg( i ) ]->toString();
					//qDebug() << "fullscale =" << fullscale( _c[ QString( "Channel %1/fullscale" ).arg( i ) ]->toString() );
					AcqrsD1_errorMessage( _instrumentID, status, msg );
					qDebug( "\tAcqrsD1_configVertical - Channel %i Status: %s", i, msg );
				}
				ViReal64 fullscale, offset;
				ViInt32 coupling,bandwidth;
				status = AcqrsD1_getVertical( _instrumentID, i, &fullscale, &offset, &coupling, &bandwidth );
				if ( status != 0 ) {
					ViChar msg[ 500 ];
					AcqrsD1_errorMessage( _instrumentID, status, msg );
					qDebug( "\tAcqrsD1_getVertical - Channel %i Status: %s", i, msg );
				}
				qDebug() << "\tChannel" << i
					<< endl << "\t\tfullscale =" << fullscale
					<< endl << "\t\toffset =" << offset
					<< endl << "\t\tcoupling =" << coupling
					<< endl << "\t\tbandwidth =" << bandwidth
					<< endl;
			}
#endif

		_verticalConfigNeeded = false;
		}
	}
	void Digitizer::scheduleHorizontalConfig() { _horizontalConfigNeeded = true; }
	void Digitizer::scheduleTriggerConfig() { _triggerConfigNeeded = true; }
	void Digitizer::scheduleVerticalConfig() { _verticalConfigNeeded = true; }

	void Digitizer::setSamplingInterval( const QString& n ) {
		qDebug() << "Digitizer::setSamplingInterval( const QString& " << n << " )";
		bool ok = false;
		double tmp = n.toDouble( &ok );
		if ( ok && tmp<1E-2 && tmp>=0.25E-9 ) {
			qDebug( "Setting samplingintervall to %e", tmp );
			_c[ "horizontal/samplinginterval" ]->setValue( tmp );
		}
	}

	void Digitizer::setLEDColor( const QString& n ) {
#ifdef HAVE_ACQIRIS
			AcqrsD1_setLEDColor( _instrumentID, LEDColors().indexOf( n ) );
#endif
	}

	QStringList Digitizer::triggersources() {
		QStringList ret;
		for ( int i=1; i<=int( channels() ); i++ )
			ret<<"Channel "+QString::number( i );
		for ( int i=1; i<=int( channels()*2 ); i++ )
			ret<<"External "+QString::number( i );
		ret.sort();
		return ret;
	}

	void Digitizer::triggersource( const QString &n ) {
		//qDebug( "\tDigitizer::triggersource( %s )", n.latin1() );
		if ( triggersources().contains( n ) ) {
			int channel=0;
			if ( n.contains( "Channel" )>0 )
				channel = n.right( -n.indexOf( " ", -1 ) ).toInt(); // XXX lastIndexOf?
			else if ( n.contains( "Extern" )>0 )
				channel = - n.right( -n.indexOf( " ", -1 ) ).toInt();
//			qDebug( "\nTrigger source will be %i\n", channel );
			_c[ "trigger/channel" ]->setValue( channel );
		}
	}
	long Digitizer::triggerSourceToPattern( int channel ) const {
#ifdef HAVE_ACQIRIS
		int nbrExtTrigsPerModule=0, nbrIntTrigsPerModule=0, nbrModules=0;
		AcqrsD1_getInstrumentInfo( _instrumentID, "NbrModulesInInstrument", &nbrModules );
		AcqrsD1_getInstrumentInfo( _instrumentID, "NbrInternalTriggers", &nbrIntTrigsPerModule );
		nbrIntTrigsPerModule = nbrIntTrigsPerModule / nbrModules;
		AcqrsD1_getInstrumentInfo( _instrumentID, "NbrExternalTriggers", &nbrExtTrigsPerModule );
		nbrExtTrigsPerModule = nbrExtTrigsPerModule / nbrModules;
		int srcpattern=0;
		long moduleNbr, inputNbr;
		if ( channel > 0 && nbrIntTrigsPerModule > 0 ) {
//			qDebug( "\t Internal" );
			moduleNbr = ( channel - 1 ) / nbrIntTrigsPerModule;
			inputNbr = ( channel - 1 ) % nbrIntTrigsPerModule;
			srcpattern = ( moduleNbr<<16 ) + ( 0x1<<inputNbr );
		} else if ( channel < 0 && nbrExtTrigsPerModule > 0 ) {
//			qDebug( "\t External" );
			channel = -channel;
			moduleNbr = ( channel - 1 ) / nbrExtTrigsPerModule;
			inputNbr = ( channel - 1 ) % nbrExtTrigsPerModule;
			srcpattern = ( moduleNbr<<16 ) + ( 0x80000000>>inputNbr );
		}
		return srcpattern;
#else
		return 0;
#endif
	}

	void Digitizer::triggerslope( const QString &n ) {
		if ( triggerslopes().contains( n ) ) {
			int set = 0;
			if ( n=="Positive" ) set = 0;
			if ( n=="Negative" ) set = 1;
			if ( n=="Out of Window" ) set = 2;
			if ( n=="Into Window" ) set = 3;
			if ( n=="HF Divide" ) set = 4;
			_c[ "trigger/slope" ]->setValue( set );
		}
	}

	void Digitizer::triggercoupling( const QString &n ) {
		if ( triggercouplings().contains( n ) ) {
			int set = 0;
			if ( n=="DC" ) set = 0;
			if ( n=="AC" ) set = 1;
			if ( n=="HF Reject" ) set = 2;
			if ( n=="DC 50Ohm" ) set = 3;
			if ( n=="AC 50Ohm" ) set = 4;
			_c[ "trigger/coupling" ]->setValue( set );
		}
	}
	double Digitizer::fullscale( QString n ) const {
		float v=5.0;
		bool ok=false;
		if ( n.contains( "mV" ) ) {
			v = n.left( n.indexOf( "mV" ) ).toFloat( &ok )/1000;
		} else if ( n.contains( "V" ) ) {
			v = n.left( n.indexOf( "V" ) ).toFloat( &ok );
		}
		if ( ok ) {
			return v;
		}
		return 5.0;
	}

	void Digitizer::triggerLevelIntern( double n ) {
		if ( _c[ "trigger/channel" ]->toInt() > 0 )
			_c[ "trigger/level" ]->setValue( n );
	}
	void Digitizer::triggerLevelExtern( double n ) {
		if ( _c[ "trigger/channel" ]->toInt() < 0 )
			_c[ "trigger/level" ]->setValue( n );
	}


}; // namespace NSAcqiris
